/**
 * synchro
 * Copyright(c) 2011 Jesus A. Domingo <jesus.domingo@gmail.com>
 * MIT Licensed
 */

/**
 * some dependencies
 */
const util = require('util'),
      events = require('events');

/**
 * the step chainer
 */
function Synchro() {
  this._async = arguments[0] || false;
  this._steps = [];
  // used only for async bundles
  this._pending = 0;
}

// for future use (listen on chains?)
util.inherits(Synchro, events.EventEmitter);

// add a function to be executed
Synchro.prototype.add = function (cb) {
  
  var step = new Step(cb),
      that = this;
  
  if (false == this._async) {
    step.on('done', function () { that.run(); });
  } else {
    step.on('done', function () { that._updatePending(); });
  }

  step.on('aborted', function () { that.abort(); });

  this._steps.push(step);
  this._pending++;

  return this;
};

// execute the chain
Synchro.prototype.run = function () {
  if (false == this._async) {
    if (this._steps.length) {
      this._steps.shift().run();
    } else {
      this.emit('done');
    }
  } else {
    var o = this._steps.shift();
    while (o) {
      o.run();
      o = this._steps.shift();
    }
    this.emit('started');
  }
};

// checks for pending async tasks
Synchro.prototype._updatePending = function () {
  this._pending--;
  if (!this._pending) {
    this.emit('done');
  }
};

// clean up the chain
Synchro.prototype.abort = function () {
  if (false == this._async) {
    throw new Exception('Asynchronous chains cannot be aborted');
  }
  if (this._steps.length) {
    this._steps.forEach(function (obj) {
      delete obj;
    });
    this._steps.length = 0;
  }
  this.emit('aborted');
};

/**
 * functions to be executed are wrapped as steps
 */
function Step(cb) {
  if ('function' != typeof cb) {
    throw new Exception('Invalid callback');
  }
  this._cb = cb;
}

util.inherits(Step, events.EventEmitter);

// execute this function/step
Step.prototype.run = function () {
  // no scope passed, maybe accept one in the future
  this._cb.call(undefined, this);
};

// emit a 'done' event to the chainer
Step.prototype.next = function () {
  this.emit('done');
};

// emit an abort to the chainer
Step.prototype.abort = function () {
  this.emit('aborted');
};

exports.create = function (sync) {
  return new Synchro(sync);
};
